#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <termios.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#ifdef linux
#include <pty.h>
#include <utmp.h>
#else
#include <libutil.h>
#endif
#include "script.h"

static void
ssh(const char *host, const char *command, const char *revision)
{
    execlp("ssh", "ssh", "-q", "-t", "-F", "ssh_config",
	   host, command, revision, (char *)NULL);
}

struct PTY {
    int master;
    int slave;
};

static int
PTY_open(struct PTY *p)
{
    return openpty(&p->master, &p->slave, NULL, NULL, NULL);
}

static void
PTY_close(struct PTY *p)
{
    close(p->master);
    if (p->slave != -1)
	close(p->slave);
}

static int
PTY_script(struct PTY *pty, FILE *fp, 
	   const char *host, const char *command, const char *revision)
{
    static const char SIGN_STR[] = "\r\nOK\r\n";
    static const size_t SIGN_LEN = sizeof(SIGN_STR) - 1;
    pid_t pid;
    int status;
    char buf[BUFSIZ];
    ssize_t size;

    if ((pid = fork()) < 0) {
	return 0;
    }
    if (pid == 0) {
	(void)close(pty->master);
	(void)fclose(fp);
        login_tty(pty->slave);
	ssh(host, command, revision);
	_exit(0);
    }
#ifdef linux
    close(pty->slave);
    pty->slave = -1;
#endif
    while ((size = read(pty->master, buf, sizeof(buf))) > 0) {
	(void)fwrite(buf, 1, size, fp);
    }
    (void)fflush(fp);
    while (wait3(&status, WNOHANG, 0) > 0) {
	continue;
    }
    if ((size = ftell(fp)) >= 0
	&& (size_t)size >= SIGN_LEN
	&& fseek(fp, -SIGN_LEN, SEEK_END) >= 0
	&& fread(buf, 1, SIGN_LEN, fp) == SIGN_LEN
	&& memcmp(buf, SIGN_STR, SIGN_LEN) == 0) {
	status = 1;
    }
    else {
	status = 2;
    }
    return status;
}

static int
scriptWithFile(FILE *fp, const char *host, const char *command,
	       const char *revision)
{
    struct PTY pty;
    int status;

    if (PTY_open(&pty) == -1)
	return 0;
    status = PTY_script(&pty, fp, host, command, revision);
    PTY_close(&pty);
    return status;
}

int
script(const char *path, const char *host, const char *command,
       const char *revision)
{
    FILE *fp;
    int status;

    if ((fp = fopen(path, "w+")) == NULL)
	return 0;
    status = scriptWithFile(fp, host, command, revision);
    (void)fclose(fp);
    return status;
}
